import numpy as np
from itertools import izip
from copy import copy
from scipy import spatial
from rlstm.common_util import unique_rows, is_nominal


def hamming_distance( x , y ):
	'''
	Measures the minimum number of substitutions required
	to change one string into the other.

	Strings are padded if not same length.
	https://en.wikipedia.org/wiki/Hamming_distance
	'''
	if len( x ) > len( y ):
		y = y + ' ' * ( len( x ) - len( y ) )
	elif len( x ) < len ( y ):
		x = x + ' ' * ( len( y ) - len( x ) )

	return sum( c1 != c2 for c1,c2 in izip( x, y ) )


def euclidean_distance( x , y ):
	return np.sqrt( np.sum( ( x - y )**2 ) )


def find_closest( x , Y , A_type ):
	'''
	Find closest vector in Y to x (but not the same).
	'''
	distfun = lambda i : euclidean_distance( x[A_type] , i[A_type] ) + \
						 hamming_distance( x[~A_type] , i[~A_type] )
	dists = np.apply_along_axis( distfun , 1 , Y )
	return np.argmin( dists[ dists > 0] )


def safe_tree_query( x , tree , k=2 , p=2 ):
	a, b = tree.query( x , k=k , p=p )
	if np.sum( a ) > 0:
		return b[-1]
	else:
		return safe_tree_query( x , tree , k=k+1 , p=p )


def munge( T , k , p , s , A_type=None ):
	'''
	MUNGE algorithm for generating arbitrary large numbers of data points
	by sampling from a non-parametric estimate of the joint distribution.

	Bucila, Caruana, Niculescu-Mizil
	https://www.cs.cornell.edu/~caruana/compression.kdd06.pdf

	Continuous attributes are assumed to be linearly scaled to [0,1].

    Args
    ----
    T : 2D array, N x A
        set of training examples,
        N is the number of samples, A is the number of attributes

    k : integer,
    	size multiplier

  	p : float (between 0 and 1),
  		defines a probability that swaps the values of attribute
  		a between examples of two sets

  	s : float,
  		standard deviation of the normal distribution to sample
  		from for continuous attributes

	A_type : 1D boolean array, A
			 True is continous , False is nominal / discrete

    Returns
    -------
    D : 2D array, (N x k) x A,
		unlabeled "munged" data set
    '''

	N , A = T.shape

	if A_type is None:
		A_type = [ True for i in range( A ) ] # default
	A_type = np.array( A_type )

	for k_i in range( k ):
		print( 'duplicating %d' % k_i )
		T_ = copy( T )
		tree = spatial.KDTree( T_ )

		for e_i in range( N ):
			e    = T_[e_i]
			e_i_ = safe_tree_query( e , tree , k=2 , p=2 )
			e_   = T_[e_i_]

			for a_i in range( A ):
				e_a  = e[a_i]
				e_a_ = e_[a_i]

				if A_type[a_i]: # continuous:
					sd = np.abs( e_a - e_a_ ) / float( s )
					if np.random.uniform() <= p:
						e_a  = np.random.normal( e_a_ , sd )
					else:
						e_a_ = np.random.normal( e_a , sd )
				else: # nominal
					if np.random.uniform() >= p:
						tmp  = e_a
						e_a  = e_a_
						e_a_ = tmp

				T_[ e_i , a_i ]  = e_a
				T_[ e_i_ , a_i ] = e_a_

		if k_i == 0:
			D = T_
		else:
			D = np.vstack( ( D , T_ ) ) # union

	return D



import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import value_and_grad
from rlstm.nn_util import sigmoid


def linear_predictions(weights, input_set, in_dim_count, out_dim_count):
    # Output_set probability of a label being true according to logistic model.
    weights = np.reshape(weights, (out_dim_count, in_dim_count))
    return np.dot(weights, input_set)


def training_loss(weights, input_set, output_set):
    # Training loss is the negative log-likelihood of the training labels.
    in_dim_count = input_set.shape[0]
    out_dim_count = output_set.shape[0]
    preds = linear_predictions(weights, input_set, in_dim_count, out_dim_count)
    loss = np.sum((output_set - preds)**2)
    return loss


def build(input_set, output_set):
    in_dim_count = input_set.shape[0]
    out_dim_count = output_set.shape[0]
    if len(input_set.shape) > 2:
        input_set = np.reshape(input_set, (in_dim_count, -1))
        output_set = np.reshape(output_set, (out_dim_count, -1))
    init_weights = npr.random((out_dim_count, in_dim_count)).flatten()
    train_iters = 10000

    loss = lambda weights: training_loss(weights, input_set, output_set)
    training_loss_and_grad = value_and_grad(loss)
    result = minimize(training_loss_and_grad, init_weights, jac=True, method='CG',
                      options={'maxiter': train_iters})
    weights = result.x
    pred_fun = lambda input_set: np.dot(
        np.reshape(weights, (out_dim_count, in_dim_count)), input_set)
    return pred_fun, weights



import autograd.numpy as np


class WeightsParser(object):

    """A helper class to index into a parameter vector."""

    def __init__(self):
        self.idxs_and_shapes = {}
        self.num_weights = 0

    def add_shape(self, name, shape):
        start = self.num_weights
        self.num_weights += np.prod(shape)
        self.idxs_and_shapes[name] = (slice(start, self.num_weights), shape)

    def get(self, vect, name):
        idxs, shape = self.idxs_and_shapes[name]
        return np.reshape(vect[idxs], shape)



import argh
from argh import arg
import cPickle
from os.path import join

import autograd.numpy as np


def normalize(X, axis=0):
    return np.divide(X.T, np.sum(X, axis=axis)).T


def toy_matrices():
    pi_mat = np.array([.5, .5, 0, 0, 0])
    trans_mat = np.array(([.5, .5, 0, 0, 0],
                          [.25, .5, .25, 0, 0],
                          [0, .25, .5, .25, 0],
                          [0, 0, .25, .5, .25],
                          [0, 0, 0, .5, .5]))
    obs_mat = np.array(([.5, .5, .5, 0, 0, 0, 0],
                        [0, .5, .5, .5, 0, 0, 0],
                        [0, 0, .5, .5, .5, 0, 0],
                        [0, 0, 0, .5, .5, .5, 0],
                        [0, 0, 0, 0, .5, .5, .5]))
    # obs_mat = np.array( ( [ 1, 0, 0, 0, 0, 0, 0 ], \
    #                       [ 0, 1, 0, 0, 0, 0, 0 ], \
    #                       [ 0, 0, 1, 0, 0, 0, 0 ], \
    #                       [ 0, 0, 0, 1, 0, 0, 0 ], \
    #                       [ 0, 0, 0, 0, 1, 0, 0 ] ) )
    return pi_mat, trans_mat, obs_mat


def spurious_matrices():
    pi_mat = np.array([.2, .2, .2, .2, .2])
    trans_mat = np.array(([.2, .2, .2, .2, .2],
                          [.2, .2, .2, .2, .2],
                          [.2, .2, .2, .2, .2],
                          [.2, .2, .2, .2, .2],
                          [.2, .2, .2, .2, .2]))
    obs_mat = np.array(([.5, .5, .5, 0, 0, 0, 0],
                        [0, .5, .5, .5, 0, 0, 0],
                        [0, 0, .5, .5, .5, 0, 0],
                        [0, 0, 0, .5, .5, .5, 0],
                        [0, 0, 0, 0, .5, .5, .5]))
    return pi_mat, trans_mat, obs_mat


def create_toy_data(data_count, time_count, bias_mat=np.array([15, 15]),
                    weight_mat=np.array(([10, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0, 0],
                                         [0, 10, 0, 0, 0, 0, 0, 10, 0, 0, 0, 0]))):
    # N is the number of samples and T is the length of the series.
    # Store data in DxTxN array because we'll want to slice by samples.

    # weight_mat is Ox(K+D)-dimensional.
    # bias_mat is O-dimensional (determines the proportion of 0s and 1s in
    # binary output).

    # Define the HMM:
    #   K is the number of states
    #   trans_mat(k,k') is the probability that we transition to k' from k (KxK-dimensional)
    # obs_mat(k,d) is the probability of 1 in dimension d in state k
    # (KxD-dimensional)
    state_count = 5  # number of states K (default = 5)
    # observation dimension D (default=7)
    dim_count = np.shape(weight_mat)[1] - state_count
    out_count = np.shape(weight_mat)[0]  # output dimension O (default=2)

    pi_mat, trans_mat, obs_mat = toy_matrices()

    # Create the sequences
    obs_set = np.zeros((dim_count, time_count, data_count), dtype=int)
    out_set = np.zeros((out_count, time_count, data_count), dtype=int)
    state_set = np.zeros((state_count, time_count, data_count), dtype=int)
    for data_iter in range(data_count):
        for time_iter in range(time_count):
            if time_iter == 0:
                state = np.random.multinomial(1, pi_mat)
                state_set[:, 0, data_iter] = state
            else:
                tvec = np.dot(
                    state_set[:, time_iter - 1, data_iter], trans_mat)
                state = np.random.multinomial(1, tvec)
                state_set[:, time_iter, data_iter] = state
    for data_iter in range(data_count):
        for time_iter in range(time_count):
            obs_vec = np.dot(state_set[:, time_iter, data_iter], obs_mat)
            obs = np.random.binomial(1, obs_vec)
            obs_set[:, time_iter, data_iter] = obs
            in_vec = np.hstack((state_set[:, time_iter, data_iter],
                                obs_set[:, time_iter, data_iter]))
            out_vec = 1 / \
                (1 + np.exp(-1 * (np.dot(weight_mat, in_vec) - bias_mat)))
            out = np.random.binomial(1, out_vec)
            out_set[:, time_iter, data_iter] = out

    # collect the outputs
    param_set = {'state_set': state_set,
                 'pi_mat': pi_mat,
                 'trans_mat': trans_mat,
                 'obs_mat': obs_mat,
                 'weight_mat': weight_mat,
                 'bias_mat': bias_mat}

    # output
    return obs_set, out_set, param_set


def create_2hmm_toy_data(data_count, time_count, bias_mat=np.array([15, 15]),
                         weight_mat=np.array(([10, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0, 0],
                                              [0, 10, 0, 0, 0, 0, 0, 10, 0, 0, 0, 0]))):
    # Define two independent HMMs. Both will help generate the observations but one
    # one of them will generate the output. The HMM not modelling output is
    # basically noise.

    state_count = 5  # number of states K (default = 5)
    # observation dimension D (default=7)
    dim_count = np.shape(weight_mat)[1] - state_count
    out_count = np.shape(weight_mat)[0]  # output dimension O (default=2)

    pi_mat_1, trans_mat_1, obs_mat_1 = toy_matrices()  # hmm1
    pi_mat_2, trans_mat_2, obs_mat_2 = spurious_matrices()  # hmm2

    # store results here (int because only integers)
    obs_set = np.zeros((dim_count * 2, time_count, data_count), dtype=int)
    out_set = np.zeros((out_count, time_count, data_count), dtype=int)
    state_set_1 = np.zeros(
        (state_count, time_count, data_count), dtype=int)  # hmm1
    state_set_2 = np.zeros(
        (state_count, time_count, data_count), dtype=int)  # hmm2

    # create state sequences for both
    for data_iter in range(data_count):
        for time_iter in range(time_count):
            if time_iter == 0:
                state_1 = np.random.multinomial(1, pi_mat_1)
                state_2 = np.random.multinomial(1, pi_mat_2)
                state_set_1[:, 0, data_iter] = state_1
                state_set_2[:, 0, data_iter] = state_2
            else:
                tvec_1 = np.dot(
                    state_set_1[:, time_iter - 1, data_iter], trans_mat_1)
                tvec_2 = np.dot(
                    state_set_2[:, time_iter - 1, data_iter], trans_mat_2)
                state_1 = np.random.multinomial(1, tvec_1)
                state_2 = np.random.multinomial(1, tvec_2)
                state_set_1[:, time_iter, data_iter] = state_1
                state_set_2[:, time_iter, data_iter] = state_2

    # generate obs and out matrices
    for data_iter in range(data_count):
        for time_iter in range(time_count):
            # obs
            obs_vec_1 = np.dot(state_set_1[:, time_iter, data_iter], obs_mat_1)
            obs_vec_2 = np.dot(state_set_2[:, time_iter, data_iter], obs_mat_2)
            obs_1 = np.random.binomial(1, obs_vec_1)
            obs_2 = np.random.binomial(1, obs_vec_2)
            obs = np.hstack((obs_1, obs_2))  # concat together
            obs_set[:, time_iter, data_iter] = obs

            # out
            in_vec = np.hstack((state_set_1[:, time_iter, data_iter],  # only use hmm1 to make output
                                obs_set[:, time_iter, data_iter][:dim_count]))
            out_vec = 1 / \
                (1 + np.exp(-1 * (np.dot(weight_mat, in_vec) - bias_mat)))
            out = np.random.binomial(1, out_vec)
            out_set[:, time_iter, data_iter] = out

    # collect the outputs
    param_set = {
        'state_set_1': state_set_1,
        'pi_mat_1': pi_mat_1,
        'trans_mat_1': trans_mat_1,
        'obs_mat_1': obs_mat_1,
        'state_set_2': state_set_2,
        'pi_mat_2': pi_mat_2,
        'trans_mat_2': trans_mat_2,
        'obs_mat_2': obs_mat_2,
        'weight_mat': weight_mat,
        'bias_mat': bias_mat
    }

    return obs_set, out_set, param_set


@arg('data_type', help='select from toy_obs_only, toy_obs_and_out, toy_two_hmm')
@arg('-tr', '--train_count', default=20, help='number of data groups in train set')
@arg('-te', '--test_count', default=10, help='number of data groups in test set')
@arg('-va', '--valid_count', default=0, help='number of data groups in validation set')
@arg('-t', '--time_count', default=50, help='number of time indices per group')
@arg('-s', '--save_path', help='directory to save numpy files')
def output_toy_data(data_type,
                    train_count=20,
                    test_count=10,
                    valid_count=0,
                    time_count=50,
                    save_path=None):
    ''' Runs create_toy_data or create_2hmm_toy_data and saves
        it into the X_train, y_train, ... style

        Args
        ----
        data_type : string
                    Choose from toy_obs_only, toy_obs_and_out, toy_two_hmm
        train_count : int
                      Number of groups in training set
        test_count : int
                     Number of groups in testing set
        valid_count : int
                      Number of groups in validation set
        time_count : int
                     Number of time steps per group
        save_path : string / None
                    Where to save the outputs

        Returns
        -------
        X_train : np array
        y_train : np array
        X_test : np array
        y_test : np array
        X_valid : np array
        y_valid : np array

    '''
    data_count = train_count + test_count + valid_count

    if data_type == 'toy_obs_only':
        bias_mat = np.array([5, 5])
        weight_mat = np.array(([10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                               [0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]))
        obs_set, out_set, param_set = \
            create_toy_data(data_count, time_count, bias_mat, weight_mat)
    elif data_type == 'toy_obs_and_out':
        bias_mat = np.array([15, 15])
        weight_mat = np.array(([10, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0, 0],
                               [0, 10, 0, 0, 0, 0, 0, 10, 0, 0, 0, 0]))
        obs_set, out_set, param_set = \
            create_toy_data(data_count, time_count, bias_mat, weight_mat)
    elif data_type == 'toy_two_hmm':
        bias_mat = np.array([15, 15])
        weight_mat = np.array(([10, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0, 0],
                               [0, 10, 0, 0, 0, 0, 0, 10, 0, 0, 0, 0]))
        obs_set, out_set, param_set = \
            create_2hmm_toy_data(data_count, time_count, bias_mat, weight_mat)

    if save_path is None:
        save_path = join('datasets', data_type)

    np.save(join(save_path, 'obs_set.npy'), obs_set)
    np.save(join(save_path, 'out_set.npy'), out_set)
    with open(join(save_path, 'param_set.pkl'), 'wb') as fp:
        cPickle.dump(param_set, fp)

    # do a train_test_split by hand
    train_obs_set = obs_set[:, :, :train_count]
    train_out_set = out_set[:, :, :train_count]
    test_obs_set = obs_set[:, :, train_count:train_count+test_count]
    test_out_set = out_set[:, :, train_count:train_count+test_count]
    valid_obs_set = obs_set[:, :, train_count+test_count:data_count]
    valid_out_set = out_set[:, :, train_count+test_count:data_count]

    np.save(join(save_path, 'X_train.npy'), train_obs_set)
    np.save(join(save_path, 'y_train.npy'), train_out_set)
    np.save(join(save_path, 'X_test.npy'), test_obs_set)
    np.save(join(save_path, 'y_test.npy'), test_out_set)
    np.save(join(save_path, 'X_valid.npy'), valid_obs_set)
    np.save(join(save_path, 'y_valid.npy'), valid_out_set)

if __name__ == '__main__':
    argh.dispatch_command(output_toy_data)




from __future__ import absolute_import, print_function

from builtins import range
from copy import copy

import autograd.numpy as np
from autograd import value_and_grad
from autograd.scipy.misc import logsumexp

from rlstm.weights_parser import WeightsParser
from rlstm.nn_util import sigmoid, add_bias_and_multiply
from rlstm.common_util import safe_log

def update(
        curr_input,
        prev_hiddens,
        update_x_weights,
        update_h_weights,
        reset_x_weights,
        reset_h_weights,
        thidden_x_weights,
        thidden_h_weights):

    update = sigmoid(add_bias_and_multiply(update_x_weights, curr_input) +
                     np.dot(prev_hiddens, update_h_weights))
    reset = sigmoid(add_bias_and_multiply(reset_x_weights, curr_input) +
                    np.dot(prev_hiddens, reset_h_weights))
    thiddens = np.tanh(add_bias_and_multiply(thidden_x_weights, curr_input) +
                       np.dot(reset * prev_hiddens, thidden_h_weights))
    hiddens = (1 - update) * prev_hiddens + update * thiddens
    return hiddens


def hiddens_to_output_probs(hiddens, output_h_weights):
    # binary case, each a prob
    output = sigmoid(np.dot(hiddens, output_h_weights))
    # output - logsumexp( output ) # normalize log-probs (multinomial case)
    # <-- softmax
    return safe_log(output), safe_log(1 - output)


def build(input_count, state_count, output_count):
    # the input count is just the number of input dimensions, we
    # assume that any/all 1-hot encoding, etc. has been done already.
    # the output count is the *number* of *binary* outputs (no
    # multinomial or continuous support at the moment)

    parser = WeightsParser()
    parser.add_shape('init_hiddens', (1, state_count))
    parser.add_shape('update_x_weights', (input_count + 1, state_count))
    parser.add_shape('update_h_weights', (state_count, state_count))
    parser.add_shape('reset_x_weights', (input_count + 1, state_count))
    parser.add_shape('reset_h_weights', (state_count, state_count))
    parser.add_shape('thidden_x_weights', (input_count + 1, state_count))
    parser.add_shape('thidden_h_weights', (state_count, state_count))
    parser.add_shape('output_h_weights', (state_count, output_count))

    def outputs(weights, input_set):
        update_x_weights = parser.get(weights, 'update_x_weights')
        update_h_weights = parser.get(weights, 'update_h_weights')
        reset_x_weights = parser.get(weights, 'reset_x_weights')
        reset_h_weights = parser.get(weights, 'reset_h_weights')
        thidden_x_weights = parser.get(weights, 'thidden_x_weights')
        thidden_h_weights = parser.get(weights, 'thidden_h_weights')
        output_h_weights = parser.get(weights, 'output_h_weights')

        data_count = input_set.shape[2]
        time_count = input_set.shape[1]

        # the output set of dimension ( output_count, time_count, data_count )
        # output_set = np.zeros( ( output_count, time_count, data_count ) )
        for data_iter in range(data_count):
            hiddens = copy(parser.get(weights, 'init_hiddens'))
            for time_iter in range(time_count):
                hiddens = update(np.expand_dims(input_set[:, time_iter, data_iter], axis=0),
                                 hiddens, update_x_weights, update_h_weights,
                                 reset_x_weights, reset_h_weights,
                                 thidden_x_weights, thidden_h_weights)

                # output_set[:,time_iter,data_iter] = \
                #    hiddens_to_output_probs( hiddens, output_h_weights )
                # more convoluted way with hstack and vstack because
                # autograd does not support assigning with indices
                out_vec1, out_vec0 = hiddens_to_output_probs(
                    hiddens, output_h_weights)
                out_vec1 = np.transpose(out_vec1)
                out_vec0 = np.transpose(out_vec0)
                if time_iter == 0:
                    out_mat1 = out_vec1
                    out_mat0 = out_vec0
                else:
                    out_mat1 = np.hstack((out_mat1, out_vec1))
                    out_mat0 = np.hstack((out_mat0, out_vec0))

            if data_iter == 0:
                output_set1 = out_mat1
                output_set0 = out_mat0
            else:
                output_set1 = np.dstack((output_set1, out_mat1))
                output_set0 = np.dstack((output_set0, out_mat0))
        return output_set1, output_set0

    def prediction(weights, input_set):
        out, _ = outputs(weights, input_set)
        return out

    def log_likelihood(weights, input_set, output_set):
        lprobs1, lprobs0 = outputs(weights, input_set)
        lprobs = output_set * lprobs1 + (1 - output_set) * lprobs0
        ll = np.sum(lprobs)
        return ll

    return outputs, log_likelihood, parser.num_weights
